from json import loads, dumps
from handshake.services.DBService.models.config_base import ConfigBase
from handshake.services.DBService import DB_VERSION
from handshake.services.DBService.models.result_base import RunBase
from handshake.services.DBService.migrator import migration
from tortoise import Tortoise, connections
from handshake.services.DBService.shared import db_path
from pathlib import Path
from typing import Optional, Union, TypedDict, Dict
from loguru import logger
from handshake.services.SchedularService.constants import (
    writtenAttachmentFolderName,
)
from sys import stderr
from click import Context
from click.core import ParameterSource


models = ["handshake.services.DBService.models"]


class VersionFile(TypedDict):
    browser_download_url: str


def handshake_meta() -> Dict["0", VersionFile]:
    version_file = Path(__file__).parent.parent.parent / ".version"
    return loads(version_file.read_text())


class DashboardMetaData(TypedDict):
    version: str
    browser_download_url_for_dashboard: str


def handshake_meta_dashboard() -> Union[DashboardMetaData, False]:
    version_file = Path(__file__).parent.parent.parent / "dashboard-meta.json"
    return False if not version_file.exists() else loads(version_file.read_text())


def save_handshake_meta_dashboard(**kwargs):
    version_file = Path(__file__).parent.parent.parent / "dashboard-meta.json"
    loaded = handshake_meta_dashboard() or {}
    loaded.update(kwargs)
    version_file.write_text(dumps(loaded))


def attachment_folder(provided_db_path: Path, *args):
    to_path = provided_db_path.parent / writtenAttachmentFolderName
    if args:
        for arg in args:
            to_path /= str(arg)
    return to_path


async def close_connection():
    await connections.close_all()
    # waiting for the logs to be sent or saved
    await logger.complete()


async def init_tortoise_orm(
    force_db_path: Optional[Union[Path, str]] = None,
    migrate: bool = False,
    close_it: bool = False,
    init_script: bool = False,
    config_path: Optional[Union[Path, str]] = None,
    avoid_config: Optional[bool] = False,
):
    chosen = force_db_path if force_db_path else db_path()
    force_init_scripts = not chosen.exists()
    # migrator is called here
    if migrate:
        migration(chosen)

    # creating a connection
    await Tortoise.init(
        db_url=r"{}".format(f"sqlite://{chosen}"),
        modules={"models": models},
    )
    # generating schemas
    await Tortoise.generate_schemas()

    test = TestConfigManager(chosen, config_path)
    # we run the init scripts for the newly created db
    await test.sync(init_script or force_init_scripts, avoid_config)

    if close_it:
        await close_connection()


async def create_run(projectName: str) -> str:
    test_id = str((await RunBase.create(projectName=projectName)).testID)
    return test_id


class TestConfigManager:
    path: Path

    def __init__(
        self,
        test_result_db: Optional[Path] = None,
        config_path: Optional[Union[str, Path]] = None,
    ):
        # default file
        # enhancement: Allow user to provide the path for the config file

        self.db_path = test_result_db
        self.pick_cfg_pth(config_path)
        if test_result_db:
            attachment_folder(self.db_path).mkdir(exist_ok=True)

    def pick_cfg_pth(self, config_path: str):
        prefs = [
            Path(config_path) if config_path else None,
            Path.cwd(),
            self.db_path.parent if self.db_path else None,
        ]
        for pref in prefs:
            if not pref:
                continue
            check = pref / "handshake.json"
            self.path = check
            if check.exists():
                break

    async def sync(
        self, init_script: bool = False, avoid_config: Optional[bool] = False
    ):
        if init_script:
            await connections.get("default").execute_script(
                f"""
            INSERT OR IGNORE INTO configbase("key", "value", "readonly") VALUES('MAX_RUNS_PER_PROJECT', '10', '0');
            INSERT OR IGNORE INTO configbase("key", "value", "readonly") VALUES('RESET_FIX_TEST_RUN', '', '1');
            INSERT OR IGNORE INTO configbase("key", "value", "readonly") VALUES('VERSION', '{DB_VERSION}', '1');
            INSERT OR IGNORE INTO configbase("key", "value", "readonly") VALUES('RECENTLY_DELETED', '0', '1');
            """,
            )
        if avoid_config:
            return None

        if not self.path.exists():
            logger.info(
                "Missing handshakes.json.,"
                " creating one at {}. we would be referring this file from now on when ran from this path.",
                self.path.parent,
            )
            return await self.save_to_file()
        await self.import_things()
        return None

    async def save_to_file(self):
        values = dict(
            zip(
                await ConfigBase.filter(readonly=False).values_list("key", flat=True),
                await ConfigBase.filter(readonly=False).values_list("value", flat=True),
            )
        )
        schema = {
            "$schema": "https://raw.githubusercontent.com/RahulARanger/handshake/"
            "refs/heads/master/handshake-config.schema.json",
            "MAX_RUNS_PER_PROJECT": int(values.get("MAX_RUNS_PER_PROJECT", 100)),
        }

        return self.path.write_text(dumps(schema, indent=4))

    async def import_things(self):
        expect_on = await ConfigBase.filter(readonly=False)
        to_save = []
        hard_save = False
        refer_from = loads(self.path.read_text())
        for record in expect_on:
            if record.key not in refer_from:
                # since some keys are missing, we are going to save them
                hard_save = True
                continue

            # make sure to understand what was provided and convert it accordingly.
            value = refer_from[record.key]
            match record.key:
                case "MAX_RUNS_PER_PROJECT":
                    value = str(refer_from[record.key])

            record.value = value
            to_save.append(record)
        to_save and await ConfigBase.bulk_update(to_save, ("value",), 100)
        if hard_save:
            logger.info(
                "Observed some of the keys are missing in handshakes.json, saving it"
            )
            await self.save_to_file()

    def get_config_for_command(self, command_name: str, quiet: bool = True):
        if self.path.exists():
            content = loads(self.path.read_text()).get("COMMANDS", {})
            value = content.get(command_name, False)
            if value:
                return value

        not quiet and logger.info(
            "You can save the configuration for this command: {} inside the handshake.json",
            command_name,
        )
        return {}


def log_less():
    logger.remove(0)
    logger.add(stderr, level="INFO")


def decide_value(context: Context, key: str, saved: dict, value_given):
    if context.get_parameter_source(
        key.lower()
    ) == ParameterSource.DEFAULT and saved.get(key):
        return saved[key]
    return value_given



from ansitohtml.parser import Parser
from loguru import logger
from typing import Optional
from abc import ABC, abstractmethod
from handshake.services.DBService.models.attachmentBase import AssertBase
from handshake.services.DBService.models.result_base import (
    SuiteBase,
    SuiteType,
    SessionBase,
)
from handshake.services.DBService.models.static_base import StaticBase
from handshake.services.SchedularService.refer_types import (
    SubSetOfRunBaseRequiredForProjectExport,
)
from handshake.services.SchedularService.constants import JobType, exportExportFileName
from asyncio import TaskGroup
from tortoise import BaseDBAsyncClient
from tortoise.connection import connections
from tortoise.expressions import Q, RawSQL
from asyncio import gather


class Exporter(ABC):
    connection: BaseDBAsyncClient

    def __init__(
        self,
        dev_run: bool = False,
    ):
        self.converter = Parser()
        self.dev_run = dev_run
        self.export_mode = dict(json=False, excel=False)

    def convert_from_ansi_to_html(self, refer_from: dict, key: str):
        refer_from[key] = self.converter.parse(refer_from[key])

    async def start_exporting(
        self,
        run_id: Optional[str] = None,
        skip_project_summary: bool = False,
        skip_prep: bool = False,
    ):
        self.connection: BaseDBAsyncClient = connections.get("default")
        if not skip_prep:
            self.prepare()
        await self.export_runs_page(run_id, skip_project_summary)
        self.completed()

    @abstractmethod
    def completed(self): ...

    @abstractmethod
    def prepare(self): ...

    async def export_runs_page(
        self, run_id: Optional[str] = None, skip_project_summary=False
    ):
        logger.debug("Exporting Runs Page...")
        path = (
            f"'/api/Attachments/' || rb.testID || '/{exportExportFileName}'"
            if self.dev_run
            else f"'/Attachments/' || rb.testID || '/{exportExportFileName}'"
        )
        extra_join_query = f"and rb.testID = '{run_id}'" if run_id else ""

        async with TaskGroup() as exporter:
            runs = []
            projects = {}

            for row in (
                await self.connection.execute_query(
                    f"""
select rb.*, cb.*,
rank() over (order by rb.ended desc) as timelineIndex,
rank() over (partition by projectName order by rb.ended desc) as projectIndex,
IIF(tb.type <> '', {path}, '') as excelExport
from RUNBASE rb
join testconfigbase cb 
on rb.testID = cb.test_id {extra_join_query}
left join taskbase tb
on tb.test_id = rb.testID and tb.type = '{JobType.EXPORT_EXCEL}'
WHERE rb.ended <> '' order by rb.started;
-- note for: projectIndex and timelineIndex, latest -> oldest => 0 - ...
    """
                )
            )[-1]:
                run = dict(row)
                test_run = SubSetOfRunBaseRequiredForProjectExport.model_validate(run)
                runs.append(run)

                logger.info(
                    "Exporting runs page for {} - {}",
                    test_run.projectName,
                    test_run.testID,
                )

                exporter.create_task(
                    self.export_test_run_summary(test_run.testID, run),
                    name=f"export-test-run-summary-{test_run.testID}",
                )

                if not skip_project_summary:
                    projects[test_run.projectName] = projects.get(
                        test_run.projectName, []
                    )
                    projects[test_run.projectName].append(
                        dict(
                            testID=test_run.testID,
                            passed=test_run.passed,
                            failed=test_run.failed,
                            skipped=test_run.skipped,
                            xfailed=test_run.xfailed,
                            xpassed=test_run.xpassed,
                            tests=test_run.tests,
                            passedSuites=test_run.passedSuites,
                            failedSuites=test_run.failedSuites,
                            skippedSuites=test_run.skippedSuites,
                            xfailedSuites=test_run.xfailedSuites,
                            xpassedSuites=test_run.xpassedSuites,
                            suites=test_run.suites,
                            duration=test_run.duration,
                        )
                    )

                exporter.create_task(
                    self.export_run_page(test_run.testID),
                    name="export-more-for-run-page",
                )

            if not skip_project_summary:
                exporter.create_task(
                    self.export_project_summary(runs, projects),
                    name="export-project-summary",
                )

            logger.info("Exported Runs Page!")

    @abstractmethod
    def export_test_run_summary(self, testID: str, results): ...

    async def export_run_page(self, run_id: str, skip_recent_suites: bool = False):
        await gather(
            self.export_overview_page(run_id, skip_recent_suites),
            self.export_all_suites(run_id),
        )

    @abstractmethod
    async def export_project_summary(self, run_feed, projects_feed): ...

    async def export_overview_page(self, run_id: str, skip_recent_suites: bool = False):
        if not skip_recent_suites:
            recent_suites = await (
                SuiteBase.filter(
                    Q(session__test_id=run_id) & Q(suiteType=SuiteType.SUITE)
                )
                .order_by("-started")
                .limit(6)
                .annotate(
                    # numberOfErrors=RawSQL("json_array_length(errors)"),
                    id=RawSQL("suiteID"),
                    s=RawSQL("suitebase.started"),
                )
                .values(
                    "title",
                    "passed",
                    "failed",
                    "skipped",
                    "xfailed",
                    "xpassed",
                    "duration",
                    suiteID="id",
                    started="s",
                )
            )
        else:
            recent_suites = []

        aggregated = (
            await SuiteBase.filter(session__test_id=run_id)
            .annotate(
                sessions=RawSQL("count(DISTINCT session_id)"),
                files=RawSQL("count(DISTINCT file)"),
            )
            .only("sessions", "files")
            .first()
            .values("sessions", "files")
        )

        platforms = (
            await SessionBase.filter(test_id=run_id)
            .only("entityName", "entityVersion", "simplified")
            .distinct()
            .values("entityName", "entityVersion", "simplified")
        )

        await self.export_overview_of_test_run(
            run_id,
            dict(
                recentSuites=recent_suites,
                aggregated=aggregated,
                platforms=platforms,
            ),
        ),

    @abstractmethod
    async def export_overview_of_test_run(self, run_id: str, summary): ...

    async def export_all_suites(self, run_id: str, export_suite_wise: bool = True):
        all_suites = await (
            SuiteBase.filter(Q(session__test_id=run_id) & Q(suiteType=SuiteType.SUITE))
            .order_by("started")
            .prefetch_related("rollup")
            .annotate(
                numberOfErrors=RawSQL("json_array_length(errors)"),
                id=RawSQL("suiteID"),
                p_id=RawSQL("parent"),
                aliasID=RawSQL(
                    "substr(suitebase.suiteID, 0, instr(suitebase.suiteID, '-'))"
                ),
                s=RawSQL("suitebase.started"),
                e=RawSQL("suitebase.ended"),
                error=RawSQL("errors ->> '[0]'"),
                total_duration=RawSQL(
                    "suitebase.duration + suitebase.setup_duration + suitebase.teardown_duration"
                ),
                parent_title=RawSQL(
                    "(select sb.title from suitebase sb where sb.suiteID = suitebase.parent)"
                ),
                nextSuite=RawSQL(
                    "(select suiteID from suitebase sb where sb.parent = suitebase.parent"
                    " and sb.session_id = suitebase.session_id and suitebase.started <= sb.started"
                    " and sb.suiteID <> suitebase.suiteID"
                    " order by sb.started limit 1)"
                ),
                prevSuite=RawSQL(
                    "(select suiteID from suitebase sb where sb.parent = suitebase.parent and sb.suiteType = 'SUITE'"
                    " and sb.session_id = suitebase.session_id and suitebase.started >= sb.started"
                    " and sb.suiteID <> suitebase.suiteID"
                    " order by sb.started limit 1)"
                ),
                hasChildSuite=RawSQL(
                    "(select count(*) from suitebase sb where sb.parent=suitebase.suiteID and sb.suiteType = 'SUITE'"
                    " and sb.suiteType='SUITE' LIMIT 1)"
                ),
            )
            .values(
                "title",
                "passed",
                "failed",
                "xpassed",
                "xfailed",
                "aliasID",
                "standing",
                "tests",
                "skipped",
                "duration",
                "file",
                "retried",
                "tags",
                "description",
                "errors",
                "error",
                "numberOfErrors",
                "hasChildSuite",
                "nextSuite",
                "prevSuite",
                "retried_later",
                "setup_duration",
                "teardown_duration",
                "total_duration",
                "parent_title",
                suiteID="id",
                parent="p_id",
                started="s",
                ended="e",
                entityName="session__entityName",
                entityVersion="session__entityVersion",
                hooks="session__hooks",
                simplified="session__simplified",
                rollup_passed="rollup__passed",
                rollup_failed="rollup__failed",
                rollup_skipped="rollup__skipped",
                rollup_xpassed="rollup__xpassed",
                rollup_xfailed="rollup__xfailed",
                rollup_tests="rollup__tests",
            )
        )
        # keep here as we would need to create directory first for the suite
        await self.export_all_suites_of_test_run(
            run_id,
            all_suites,
        )
        if export_suite_wise:
            await gather(
                *[
                    self.export_suite(run_id, str(suite))
                    for suite in await SuiteBase.filter(
                        Q(session__test_id=run_id) & Q(suiteType=SuiteType.SUITE)
                    ).values_list("suiteID", flat=True)
                ],
            )
        else:
            await self.export_suite(run_id)

    @abstractmethod
    async def export_all_suites_of_test_run(self, run_id, all_suites): ...

    @staticmethod
    def export_test_query(based_on: str):
        return (
            SuiteBase.filter(Q(parent=based_on)),
            AssertBase.filter(entity__parent=based_on),
            StaticBase.filter(entity__parent=based_on),
        )

    async def export_suite(self, run_id: str, suite_id: Optional[str] = None):
        test_query, assertion_query, attachment_query = self.export_test_query(
            suite_id if suite_id else run_id
        )
        tests = await (
            test_query.order_by("-suiteType", "started")
            .prefetch_related("rollup")
            .annotate(
                aliasID=RawSQL(
                    "substr(suitebase.suiteID, 0, instr(suitebase.suiteID, '-'))"
                ),
                numberOfErrors=RawSQL("json_array_length(errors)"),
                id=RawSQL("suiteID"),
                s=RawSQL("suitebase.started"),
                total_duration=RawSQL(
                    "suitebase.duration + suitebase.setup_duration + suitebase.teardown_duration"
                ),
                parent_title=RawSQL(
                    "(select sb.title from suitebase sb where sb.suiteID = suitebase.parent)"
                ),
                e=RawSQL("suitebase.ended"),
                error=RawSQL("errors ->> '[0]'"),
                assertions=RawSQL(
                    "(select count(ab.entity_id) from assertbase ab where ab.entity_id=suitebase.suiteID)"
                ),
            )
            .values(
                "title",
                "aliasID",
                "standing",
                "assertions",
                "duration",
                "file",
                "retried",
                "tags",
                "suiteType",
                "description",
                "errors",
                "error",
                "numberOfErrors",
                "retried_later",
                "setup_duration",
                "teardown_duration",
                "total_duration",
                "parent",
                "parent_title",
                suiteID="id",
                started="s",
                ended="e",
                hooks="session__hooks",
                rollup_passed="rollup__passed",
                rollup_failed="rollup__failed",
                rollup_skipped="rollup__skipped",
                rollup_tests="rollup__tests",
                rollup_xpassed="rollup__xpassed",
                rollup_xfailed="rollup__xfailed",
            )
        )

        assertions = (
            await assertion_query.annotate(
                id=RawSQL("entity_id"),
                raw=RawSQL("message"),
                aliasID=RawSQL(
                    "substr(assertbase.entity_id, 0, instr(assertbase.entity_id, '-'))"
                ),
            )
            .all()
            .values(
                "title",
                "message",
                "raw",
                "interval",
                "passed",
                "wait",
                "aliasID",
                entity_id="id",
            )
        )

        written_records = {}
        # TODO: Export Attachments to excel
        assertion_records = {}
        written = (
            await attachment_query.annotate(
                id=RawSQL("entity_id"),
                file=RawSQL("value"),
                url=RawSQL(
                    f"'/api/Attachments' || '/{run_id}/' || entity_id || '/' || value"
                    if self.dev_run
                    else f"'/Attachments' || '/{run_id}/' || entity_id || '/' || value"
                ),
            )
            .all()
            .values("type", "title", "description", "file", "url", entity_id="id")
        )

        for refer_from, save_in, for_records in zip(
            (written, assertions),
            (written_records, assertion_records),
            ("written", "assertions"),
        ):
            for record in refer_from:
                records = save_in.get(record["entity_id"], [])
                records.append(record)
                if for_records == "assertions":
                    self.convert_from_ansi_to_html(record, "message")
                save_in[record["entity_id"]] = records

        retried_map = {}

        _, rows = await connections.get("default").execute_query(
            "select key, value as suite, rb.tests, length, suite_id"
            " from retriedbase rb join json_each(rb.tests)"
            " join suitebase sb on suite = sb.suiteID"
            " join sessionbase ssb on ssb.sessionID = sb.session_id"
            " where ssb.test_id = ?",
            (run_id,),
        )

        for row in rows:
            retried_map[row["suite"]] = dict(row)

        await gather(
            self.export_tests(run_id, suite_id, tests),
            self.export_attachments(
                run_id, suite_id, assertion_records, written_records
            ),
            self.export_retries_map(run_id, suite_id, retried_map),
        )

    @abstractmethod
    def export_tests(self, run_id, suite_id, tests): ...

    @abstractmethod
    def export_attachments(
        self, run_id, suite_id, assertion_records, written_records
    ): ...

    @abstractmethod
    def export_retries_map(self, run_id, suite_id, retried_map): ...



from httpx import Client
from datetime import datetime
from subprocess import Popen
from typing import Union, Optional, Dict, List
from loguru import logger
from time import sleep
from sys import stdout, stderr
from handshake.services.DBService.models.types import (
    MarkTestRun,
    PydanticModalForCreatingTestRunConfigBase,
)
from threading import Lock
from concurrent.futures.thread import ThreadPoolExecutor
from concurrent.futures import Future
from httpx import Response


def to_acceptable_date_format(date: datetime):
    return date.isoformat()


class CommonReporter:
    results: str
    port: str
    url: str
    client: Client
    collector: Popen
    postman: ThreadPoolExecutor
    started: Future
    skip: bool = False
    config_path: Optional[str] = None

    def __init__(self, path: str = "TestResults", port: Union[str, int] = 6969):
        self.note = dict()
        self.set_context(False, path, port)
        self.lock_attachments = Lock()
        self.connection_established = False
        self.waiting = Lock()
        self.attachments: List[Dict] = []

    def postfix(self, fix: str):
        return f"{self.url}/{fix}"

    def create_postfix(self, fix: str):
        return self.postfix(f"create/{fix}")

    def update_postfix(self, fix: str):
        return self.postfix(f"save/{fix}")

    def ensure_mails(
        self, postman, url, note: Optional[Union[str, False]] = False, **kwargs
    ):
        try:
            if kwargs.get("json", False) and kwargs.get("append", False):
                kwargs["json"] = {
                    **kwargs["json"],
                    **{
                        refer_as: self.note.get(refer_to, "")
                        for refer_as, refer_to in kwargs["append"].items()
                    },
                }

            if kwargs.get("json", False) and kwargs.get("map_value", False):
                for change_for in kwargs["json"]:
                    before = change_for[kwargs["map_value"]]
                    change_for[kwargs["map_value"]] = self.note[before]

            for to_pop in ("append", "map_value"):
                kwargs.pop(to_pop) if to_pop in kwargs else ...

            response: Response = postman(url, **kwargs)
            if response.status_code // 200 != 1:
                logger.warning(
                    "Request failed: {}, status: {}. response: {}",
                    url,
                    response.status_code,
                    response.text,
                )

            response.raise_for_status()
            if note:
                self.note[note] = response.text
            return response
        except Exception as error:
            logger.exception(
                "Failed to send request to {} with payload: {}. due to {}",
                url,
                kwargs,
                repr(error),
            )
            return False

    def set_context(
        self,
        set_client: bool,
        path: str,
        port: Union[str, int],
        config_path: Optional[str] = None,
    ):
        self.results = path
        self.port = str(port)
        self.client = Client() if set_client else ...

        self.postman = (
            ThreadPoolExecutor(max_workers=1, thread_name_prefix="post-")
            if set_client
            else ...
        )
        self.url = f"http://127.0.0.1:{port}"
        self.config_path = config_path

    def start_collection(self, project_name: str, is_quiet: bool):
        postfix = " " + ("" if is_quiet else "-vb")
        command = (
            f'handshake run-app {project_name} "{self.results}" "{self.config_path}" -p {self.port}'
            if self.config_path
            else f'handshake run-app {project_name} "{self.results}" -p {self.port}'
        ) + postfix
        self.collector = Popen(command, shell=True, stdout=stdout, stderr=stderr)
        logger.debug("Starting handshake-server, {}", command)
        self.started = self.postman.submit(self.wait_for_connection)

    def health_connection(self) -> bool:
        return self.collector.poll() is None

    def set_skip(self, error):
        # logger.error("Skipping Handshake reports, because {}", error)
        self.skip = True

    def wait_for_connection(self, retried: int = 1, force_call: bool = False):
        stack = [retried]
        with self.waiting:
            while stack:
                to_retry = stack.pop()
                try:
                    if self.skip and not force_call:
                        return None
                    if not self.health_connection():
                        return not force_call and self.set_skip(
                            "Handshake server has closed."
                        )

                    self.client.get(
                        self.postfix(""),
                        timeout=60,
                    ).raise_for_status()
                    self.connection_established = True

                    not force_call and logger.debug(
                        "Connection established with handshake server."
                    )
                except Exception as _:
                    sleep(to_retry * 0.5)
                    if to_retry == 6:
                        return not force_call and self.set_skip(
                            "Failed to connect with handshake server, cancelling reports to send."
                        )
                    stack.append(to_retry + 1)
            return None

    def call(
        self,
        reason: str,
        post_it: bool,
        postfix: str,
        payload: Union[Dict, List],
        save_it: Optional[str] = None,
        map_value: str = None,
        append: Optional[Dict[str, str]] = None,
    ):
        logger.debug(reason)
        self.postman.submit(
            self.ensure_mails,
            self.client.post if post_it else self.client.put,
            (self.create_postfix if post_it else self.update_postfix)(postfix),
            save_it if save_it else False,
            json=payload,
            append=append,
            map_value=map_value,
        )

    def create_session(self, started: datetime):
        return self.call(
            "Creating test session",
            True,
            "Session",
            dict(started=to_acceptable_date_format(started)),
            "Session",
        )

    @property
    def session(self):
        return self.note["Session"]

    def add_run_config(self, payload: PydanticModalForCreatingTestRunConfigBase):
        return self.call(
            "Adding Configuration for your test run",
            True,
            "RunConfig",
            payload.model_dump(),
        )

    def register_test_entity(
        self, payload: Dict, save_in: str, parent: Optional[str] = None
    ):
        to_append = dict(session_id="Session")
        if parent:
            to_append["parent"] = parent

        return self.call(
            f"Registering a Test Entity: {payload['title']}",
            True,
            "Suite",
            payload,
            save_in,
            append=to_append,
        )

    def update_test_entity(
        self, payload, node_id: str, punch_in: Optional[bool] = False
    ):
        self.send_chunk_of_attachments()
        return self.call(
            f"Updating Test Entity: {node_id}",
            False,
            "PunchInSuite" if punch_in else "Suite",
            payload,
            append=dict(suiteID=node_id),
        )

    def send_chunk_of_attachments(self):
        with self.lock_attachments:
            if not self.attachments:
                return

            self.call(
                "Sending chunk of attachments",
                True,
                "Attachments",
                [*self.attachments],
                map_value="entity_id",
            )
            self.attachments.clear()

    def update_test_session(self, payload: Dict):
        return self.call(
            "Updating Test Session",
            False,
            "Session",
            payload,
            append=dict(sessionID="Session"),
        )

    def update_test_run(self, payload: MarkTestRun, force_call=False):
        if force_call:
            return self.client.put(
                self.update_postfix("Run"), json=payload.model_dump()
            )
        else:
            self.send_chunk_of_attachments()
        return self.call(
            "Updating Test Run",
            False,
            "Run",
            payload.model_dump(),
        )

    def force_wait(self):
        if not self.health_connection():
            return None

        return self.wait_for_connection(1, True)

    def close_resources(self, force_call=False):
        with self.waiting:
            if force_call:
                return self.client.post(self.postfix("bye"))
            else:
                self.send_chunk_of_attachments()
            self.postman.submit(
                self.ensure_mails, self.client.post, self.postfix("bye"), False
            )
        self.postman.shutdown(wait=True, cancel_futures=False)
        logger.complete()

        if not self.skip:
            logger.debug("Handshake Reporter has collected your reports.")
        return None



from datetime import datetime
from handshake.reporters.reporter import CommonReporter, to_acceptable_date_format
from handshake.services.DBService.models.types import (
    PydanticModalForCreatingTestRunConfigBase,
    MarkTestRun,
    AttachmentType,
    Tag,
)
from handshake.services.DBService.models.enums import (
    LogType,
    RunStatus,
    SuiteType,
    Status,
)
from handshake.reporters.markers import meta_data_mark
from pytest import Session, Item, ExitCode, TestReport
from platform import platform
from typing import Optional, Dict, Any, List
from threading import Lock
from enum import StrEnum
from _pytest.fixtures import FixtureDef, FixtureValue, SubRequest
from _pytest.nodes import Node
from pathlib import Path


def relative_from_session_parent(session: Session, take_relative_for: Path):
    if not take_relative_for.is_relative_to(session.startpath.parent):
        return None
    return take_relative_for.relative_to(session.startpath.parent)


class PointToAtPhase(StrEnum):
    CALL = "call"
    SETUP = "setup"
    TEARDOWN = "teardown"


def key(node_id: str, method: Optional[str] = "call"):
    return node_id + "-" + method


class PyTestHandshakeReporter(CommonReporter):
    def __init__(self):
        super().__init__()
        self.pointing_to: Optional[Item] = None
        self.identified_parent = dict()
        self.check_if_parents_are = Lock()
        self.mark_point = Lock()
        self.passed = 0
        self.started_at = None
        self.attachments: List[Dict[str, Any]] = []
        self.fixtures = {}
        self.func_args = {}

    def parse_config(self, session: Session):
        unregister = session.config.inicfg.get("disable_handshakes")
        if unregister:
            handshake_plugin = session.config.pluginmanager.get_plugin("handshakes")
            session.config.pluginmanager.unregister(handshake_plugin)
            return True

        rel_path = session.config.inicfg.get("save_results_in")
        port = session.config.inicfg.get("handshake_port")
        config_path = session.config.inicfg.get("save_handshake_config_dir")
        rel_to = Path(session.config.inipath.parent)

        self.set_context(
            True,
            (rel_to / rel_path).resolve() if rel_path else self.results,
            port if port else self.port,
            (rel_to / config_path).resolve() if config_path else None,
        )

    def create_session(self, started: datetime):
        super().create_session(started)
        self.started_at = started

    def put_test_config(self, config: dict[str, str | list[str]]):
        test_run_tag_prefix = "handshake_tag_"

        tags = []
        for key_desc, value in config.items():
            if not key_desc.startswith(test_run_tag_prefix):
                continue
            tags.append(Tag(label=value, desc=key_desc[len(test_run_tag_prefix) :]))

        return self.add_run_config(
            PydanticModalForCreatingTestRunConfigBase(
                framework="pytest",
                platform=platform(),
                tags=tags,
                avoidParentSuitesInCount=False,  # since we are counting packages as suites
            )
        )

    def get_key(self, node_id: str, method: Optional[str] = "call"):
        return self.note[key(node_id, method)]

    def set_key(self, node_id: str, value: str, method: Optional[str] = "call"):
        self.note[key(node_id, method)] = value

    def create_test_entity(
        self,
        item: Node,
        is_suite: bool = False,
        helper_entity: Optional[PointToAtPhase] = None,
    ):
        path_obj: Path = relative_from_session_parent(item.session, item.path)
        if path_obj is None:
            return

        if helper_entity:
            with self.mark_point:
                self.pointing_to = key(item.nodeid, helper_entity)

        path_obj = path_obj.parent if path_obj.name == "__init__.py" else path_obj
        path: str = str(path_obj)
        parent = key(item.nodeid) if helper_entity else None

        if (
            helper_entity is None
            and item.parent.nodeid
            and not Path(item.parent.path).is_dir()
        ):
            parent = key(item.parent.nodeid)
            with self.check_if_parents_are:
                create_parent = not self.identified_parent.get(
                    item.parent.nodeid, False
                )
            if create_parent:
                self.create_test_entity(item.parent, True)
                with self.check_if_parents_are:
                    self.identified_parent[item.parent.nodeid] = True

        suite_type = SuiteType.SUITE if is_suite else SuiteType.TEST
        marker = not helper_entity and item.get_closest_marker(meta_data_mark)
        meta = marker.kwargs if marker else {}
        tags = []

        for tag in item.own_markers if not helper_entity else tags:
            if tag.name == meta_data_mark:  # if it is our tag, we ignore it
                continue
            desc = ("" if not tag.args else f"args: {tag.args} and ") + (
                "kwargs: {}" if not tag.kwargs else f"kwargs: {tag.kwargs}"
            )
            tags.append(dict(label=tag.name, desc=desc))

        title = ""
        description = ""

        if hasattr(item, "callspec"):
            if hasattr(item.callspec, "id"):
                meta_obj = meta.get(item.callspec.id, dict())
                title = meta_obj.get("title")
                description = meta_obj.get("description")

        title = title or meta.get(title, item.name)
        description = description or meta.get("description", "")

        self.register_test_entity(
            dict(
                file=path,
                title=title,
                description=description,
                suiteType=helper_entity.upper() if helper_entity else suite_type,
                parent="",
                is_processing=helper_entity is not None,
                session_id="",
                tags=tags,
            ),
            key(item.nodeid, helper_entity if helper_entity else PointToAtPhase.CALL),
            parent,
        )

        if is_suite:
            return

        if hasattr(item, "fixturenames"):
            for name in item.fixturenames:
                add_here = self.fixtures.get(name, set())
                add_here.add(
                    key(
                        item.nodeid,
                        helper_entity if helper_entity else PointToAtPhase.CALL,
                    )
                )
                self.fixtures[name] = add_here

    def update_test_entity_details(
        self, report: Optional[TestReport] = None, node_id: Optional[str] = None
    ):
        with self.mark_point:
            self.pointing_to = (
                key(node_id) if node_id else key(report.nodeid, report.when)
            )
        if not report:
            return self.update_test_entity(
                dict(started=to_acceptable_date_format(datetime.now())),
                key(node_id),
                punch_in=True,
            )

        # NOTE: there are test cases that are skipped before the test is called
        # we do not have a report for the call.
        # so we need to manually update for such
        # Hack: we assume setup call as usual test
        when = (
            PointToAtPhase.CALL
            if report.skipped and report.when == PointToAtPhase.SETUP
            else report.when
        )

        outcome = (
            (Status.XPASSED if report.passed else Status.XFAILED)
            if report.when == PointToAtPhase.CALL
            and report.keywords.get("xfail", 0) == 1
            else report.outcome
        ).upper()

        payload = dict(
            duration=report.duration * 1e3,  # it is in seconds
            ended=to_acceptable_date_format(datetime.fromtimestamp(report.stop)),
            started=to_acceptable_date_format(datetime.fromtimestamp(report.start)),
            standing=outcome,
            errors=(
                [dict(name="Error", stack="", message=report.longreprtext)]
                if report.failed
                else []
            ),
        )

        if report.skipped:
            self.add_log(
                LogType.INFO,
                key(report.nodeid),
                "Skipped!",
                report.longreprtext,
                tags=[dict(label="Reason", desc=repr(report.longrepr))],
            )

        match when:
            case PointToAtPhase.CALL:
                self.passed += int(report.passed)

        self.update_test_entity(
            payload,
            key(report.nodeid, when),
        )

    def update_session(self, session: Session):
        ended = datetime.now()
        self.update_test_session(
            dict(
                ended=to_acceptable_date_format(ended),
                entityName="console",
                entityVersion="-",
                simplified="console",
                tests=session.testscollected,
                failed=session.testsfailed,
                passed=self.passed,
                skipped=(session.testscollected - (session.testsfailed + self.passed)),
                duration=(ended - self.started_at).total_seconds() * 1e3,
                hooks=0,
            )
        )

    def update_test_status(self, session: Optional[Session] = None, exitcode: int = 1):
        status: RunStatus = RunStatus.COMPLETED
        if session.shouldfail or session.shouldstop:
            status = RunStatus.EXPECTED_TO_FAIL
        elif session.exitstatus == ExitCode.INTERRUPTED:
            status = RunStatus.INTERRUPTED
        elif (
            session.exitstatus == ExitCode.USAGE_ERROR
            or session.exitstatus == ExitCode.INTERNAL_ERROR
        ):
            status = RunStatus.INTERNAL_ERROR

        force_call = status == RunStatus.INTERRUPTED

        if force_call:
            self.skip = True
            self.force_wait()

        self.update_test_run(
            MarkTestRun(exitCode=exitcode, status=status), force_call=force_call
        )
        return force_call

    def mark_suites_for_processing(self):
        return self.postman.submit(
            self.ensure_mails,
            self.client.post,
            self.create_postfix("ScheduleSuites"),
            False,
        )

    def add_log(
        self,
        log_type: LogType,
        entity_id: str,
        title: str,
        description: str,
        tags: Optional[List[Tag]] = None,
    ):
        self.attachments.append(
            dict(
                type=AttachmentType.LOG,
                entity_id=entity_id,
                description=description,
                value=dict(type=log_type),
                title=title,
                tags=tags or [],
            )
        )

    def note_fixture(self, fixturedef: FixtureDef[FixtureValue], request: SubRequest):
        if fixturedef.cached_result is None:
            return

        scope_desc = ""

        match request.scope:
            case "session":
                scope_desc = "Shared across all tests in the session."
            case "package":
                scope_desc = "Shared across tests in a package."
            case "function":
                scope_desc = "Created anew for each test function."
            case "class":
                scope_desc = "Shared among tests in a class."

        if fixturedef.cached_result and fixturedef.cached_result[-1]:
            note_desc = (
                f"{request.fixturename}(scope: {request.scope}) failed to execute, "
                f"because of this error: {repr(fixturedef.cached_result[-1][1])}"
            )
        else:
            note_desc = f"{request.fixturename}(scope: {request.scope}) passed and it gave a result {str(fixturedef.cached_result[0])}"

        fixture_def = "A fixture provides a defined, reliable and consistent context for the tests"

        save_in = relative_from_session_parent(
            request.session,
            request.session.startpath if request.scope == "session" else request.path,
        )

        for add_in in self.fixtures.get(request.fixturename, set()):
            self.add_log(
                LogType.INFO,
                entity_id=add_in,
                title=request.fixturename,
                tags=[
                    dict(  # https://docs.pytest.org/en/stable/explanation/fixtures.html
                        label="fixture",
                        desc=fixture_def,
                    ),
                    dict(label=request.scope, desc=scope_desc),
                    dict(
                        label="saved in",
                        desc=str(save_in if save_in else request.session.startpath),
                    ),
                ],
                description=note_desc,
            )

        self.fixtures.get(request.fixturename, set()).clear()

    def add_test_assertion(self, node_id: str, title: str, message: str, passed: bool):
        with self.mark_point:
            value = dict(
                entity_id=key(node_id) if node_id else self.pointing_to,
                title=title,
                value=dict(passed=passed, wait=-1, interval=-1),
                type=AttachmentType.ASSERT,
                description=message,
            )
            return self.attachments.append(value)



import sqlite3
from sqlite3 import connect, sqlite_version_info
from click import (
    group,
    argument,
    secho,
    version_option,
    pass_context,
    Context,
    Path as C_Path,
    confirm,
)
from shutil import make_archive, move, unpack_archive
from tortoise import run_async
from handshake import __version__
from handshake.services.DBService.migrator import (
    check_version,
    migration,
    revert_step_back,
    DB_VERSION,
    MigrationTrigger,
)
from handshake.services.SchedularService.start import Scheduler
from loguru import logger
from handshake.services.DBService.lifecycle import (
    close_connection,
    init_tortoise_orm,
    log_less,
    decide_value,
    TestConfigManager,
)
from handshake.services.DBService.merge import Merger
from click import option, get_current_context
from pathlib import Path
from handshake.services.DBService.shared import db_path
from typing import Optional
from handshake.Exporters.html_reporter import HTMLExporter
from asyncio import run


@group(
    name="Handshake",
    short_help="Handshake command",
    help=f"""

{'{:*^69}'.format(" Welcome to Handshake's CLI ")}

Handshake simplifies the collection and processing of your test results. The Handshake CLI helps you to interact with the stored results or with the server. so for each command it requires a <path> argument, representing the collection 
folder, a place where you could have the database or were planning to store the results [not the HTML files].

[ROOT-DIR] >> [COLLECTION_NAME] (*we need this) >> TeStReSuLtS.db & [Attachments] (for getting this).

{'{:*^69}'.format(" Glossary ")}
""",
)
@version_option(f"{__version__}, DB: {DB_VERSION}")
def handle_cli():
    pass


general_requirement = argument(
    "COLLECTION_PATH", nargs=1, type=C_Path(exists=True, dir_okay=True), required=True
)
general_but_optionally_present = argument(
    "COLLECTION_PATH", nargs=1, type=C_Path(dir_okay=True), required=True
)
config_optional_path = argument(
    "CONFIG_PATH", nargs=1, type=C_Path(dir_okay=True), required=False
)

observed_version = option(
    "--version",
    "-v",
    default="",
    type=str,
    required=False,
    show_default=True,
    help="Used Internally with the reporters,"
    " reporters will pass the version of the expected handshake server if it matches,"
    " we run the server else we terminate the execution.",
)


def break_if_mismatch(expected: str) -> bool:
    if expected:
        assert expected.strip() == __version__, (
            f"Mismatch between expected version: {expected} "
            f"and the actual version v{__version__}"
        )
    return True


@general_requirement
@handle_cli.command(
    short_help="Returns the version inside the TestResults",
    help="it would be possible the server and stored results are in different version, "
    "with this command you can clear that query and note if migration is required",
)
def db_version(collection_path):
    return check_version(path=db_path(collection_path), is_auto=True)


@handle_cli.command(
    short_help="checks the version of the sqlite3 installed in your system"
)
def check_sqlite():
    assert int(sqlite_version_info[0]) >= 3, "Required version is >= 3."
    assert (
        int(sqlite_version_info[1]) >= 38
    ), "Required Version is >= 3.38,  for supporting our sql scripts, for supporting our sql scripts"
    logger.info("your sqlite version is fine as per our need")


@general_requirement
@handle_cli.command(
    short_help="Migrates the database to the latest version as per the handshake executable.",
    help="it's a command to execute the required migration scripts, note; this command would be executed "
    "automatically whenever we run export or run-app command",
)
def migrate(collection_path: str):
    return migration(db_path(collection_path), MigrationTrigger.CLI)


@general_requirement
@handle_cli.command(
    short_help="To revert db's version one step back. USE IT WITH CAUTION",
    help="you might need to use this if you are planning to use previous version python build."
    " Make sure to understand why this command is useful and how it would impact."
    " After this command, if db version was in v8 it would not be in v7",
)
def step_back(collection_path: str):
    path_to_refer = db_path(collection_path)
    from_version = (check_version(path=path_to_refer, is_auto=True))[-1]
    if confirm(f"Do you want revert from v{from_version} to v{from_version - 1}"):
        revert_step_back(from_version, path_to_refer)


@handle_cli.command(
    short_help="Processes the collected results and even could export the test results",
    help="runs an async loop, schedules some tasks to export some your test results "
    "so you can see it in the way we need. you can pass the output directory to generate the report",
)
@option(
    "-vb",
    "--verbose",
    default=False,
    show_default=True,
    help="shows even debug logs",
    is_flag=True,
)
@option(
    "-r",
    "--reset",
    default=False,
    show_default=True,
    help="re-calculates the values for the test runs",
    is_flag=True,
)
@general_requirement
@config_optional_path
@option(
    "--inside",
    "-i",
    required=False,
    help="generates the export inside the TestResults itself",
    is_flag=True,
    default=False,
    show_default=True,
)
@option(
    "--output-folder",
    "-o",
    help="generates the export at this desired place",
    type=C_Path(dir_okay=True, writable=True),
    required=False,
)
@option(
    "--export_mode",
    "-e",
    help="generates either json/html export",
    type=str,
    required=False,
    default="json",
)
@option(
    "--include-excel",
    "-xl",
    help="generates excel export of the test run",
    required=False,
    is_flag=True,
    default=False,
    show_default=True,
)
@option(
    "--dev",
    "-d",
    required=False,
    help="This is used only for testing purposes, please do not use this flag",
    is_flag=True,
    default=False,
    show_default=False,
)
def export(
    collection_path,
    reset: bool = False,
    config_path: Optional[str] = None,
    output_folder: str = None,
    verbose: bool = False,
    dev: bool = False,
    inside=False,
    export_mode: str = "json",
    include_excel: bool = False,
):
    q = not (verbose or dev)
    if q:
        log_less()

    if not Path(collection_path).is_dir():
        raise NotADirectoryError(collection_path)

    refer_from_here = TestConfigManager(config_path=config_path).get_config_for_command(
        "EXPORT", q
    )
    context = get_current_context()
    export_mode = decide_value(context, "EXPORT_MODE", refer_from_here, export_mode)
    out_folder = decide_value(context, "OUTPUT_FOLDER", refer_from_here, output_folder)

    if export_mode.lower() == "html" and out_folder is None:
        logger.error(
            "HTML export requires an output folder, please provide it using --out or -o"
        )
        return

    scheduler = Scheduler(
        collection_path,
        out_folder,
        reset,
        inside,
        dev,
        export_mode,
        decide_value(context, "INCLUDE_EXCEL", refer_from_here, include_excel),
    )
    try:
        run(scheduler.start(config_path))
    except (KeyboardInterrupt, SystemExit):
        logger.warning("Scheduler terminated explicitly...")
        run(close_connection())


@handle_cli.command(
    short_help="zips your TestResults in a tar file",
    help="it zips your TestResults in a single tar file with bz2 compression level."
    " it stores only inside the provided TestResults folder but not the TestResults folder itself.",
)
@general_requirement
@option(
    "--out",
    "-o",
    help="Saves the zipped file inside this folder path",
    type=C_Path(dir_okay=True),
    required=False,
    default=None,
)
def zip_results(collection_path, out=None):
    collection = Path(collection_path)
    output_folder = Path(out).resolve() if out else Path.cwd()
    output_folder.mkdir(exist_ok=True)

    logger.info(f"compressing TestResults located at: {collection}")
    file_name = collection.stem + ".tar.bz2"
    before = (output_folder / file_name).exists()
    if before:
        logger.error(
            f"output file already exists, Please remove or rename the file at {output_folder / file_name}",
        )

    else:
        make_archive(collection.stem, "bztar", collection)
        if out:
            move(file_name, output_folder)
        logger.info(f"Done, located at {output_folder / file_name}")


@handle_cli.command(
    short_help="extracts zipped (.bz2) TestResults into a provided folder",
    help="extracts zipped TestResults into a provided folder. "
    "we assume that the results were zipped in a tar file with bz2 compression",
)
@argument(
    "file",
    type=C_Path(file_okay=True, readable=True, exists=True),
    required=True,
)
@option(
    "--out",
    "-o",
    help="Extracts results inside this folder",
    type=C_Path(dir_okay=True, writable=True),
    required=False,
)
def extract_results(file: str, out: str):
    output_folder = Path(out if out else file.split(".")[0])
    logger.info(f"de-compressing {file}")
    unpack_archive(file, output_folder, "bztar")
    logger.info(f"Done, located at {output_folder}")


@handle_cli.command(
    short_help="merges all the provided results into the single one. Simple Merge",
    help="It's a simple merge, it takes in either zip file of your TestResults or a folder itself "
    "and then it takes a copy, migrates if required and then inserts data into the output database."
    " Please note we would also run migration scripts on the output TestResults. Temp folders are then deleted.",
)
@argument(
    "output",
    type=C_Path(dir_okay=True, writable=True, exists=False),
    required=True,
)
@option(
    "-m",
    "--merge-with",
    required=True,
    multiple=True,
    help="provide the compressed path of the results folder",
)
def merge(output, merge_with):
    merger = Merger(output)
    merger.start(merge_with)


@handle_cli.command(
    short_help="downloads the dashboard build",
    help="downloads the build for the dashboard it is required for allowing the html export,"
    " we execute this command to forcefully download the build, it is not required to run this manually.",
)
def download_build():
    run(HTMLExporter.download_zip())


@handle_cli.command(
    help="returns the version of the handshake", short_help="example: 1.0.0"
)
def v():
    secho(__version__)


@handle_cli.command(
    short_help="does the required setup to store your test results in this folder",
    help="""
Configures TestResults folder with the provided folder name at your cwd. 
Example: handshake init TestResults, at your cwd: x,\n
then it creates x -> TestResults -> TeStReSuLtS.sb and x -> TestResults -> Attachments
and x -> handshakes.json
    """,
)
@general_but_optionally_present
@config_optional_path
def init(collection_path, config_path=None):
    saved_db_path = db_path(collection_path)

    set_default_first = not Path(collection_path).exists()
    if set_default_first:
        Path(collection_path).mkdir(exist_ok=True)

    run_async(
        init_tortoise_orm(
            saved_db_path,
            True,
            close_it=True,
            init_script=True,
            config_path=config_path,
        )
    )


@handle_cli.group(
    name="faq",
    short_help="Commands that will/(try to) fetch mostly asked info. from db",
    help="you can query using db following the subcommands, which are created to provide mostly asked info. from db",
)
@version_option(str(DB_VERSION), prog_name="handshake-db")
@general_requirement
def faq(collection_path: str):
    if not Path(db_path(collection_path)).exists():
        raise FileNotFoundError(
            f"db path was not found in the collections: {collection_path}"
        )


@faq.command(short_help="fetches the results from the sqlite database")
@argument(
    "q",
    default=False,
    type=str,
)
@pass_context
def query(ctx: Context, q: str, collection_path: Optional[str] = None):
    db_file = db_path(Path(collection_path or ctx.parent.params["collection_path"]))
    try:
        import tabulate
    except ImportError:
        return logger.error(
            "could not execute the provided query, Please install this package by"
            ' pip install "handshakes[print-tables]"'
        )
    if not q.lower().startswith("select"):
        return logger.warning(
            "we do not support & recommend modifying the values through this command, only select commands are allowed"
        )

    with connect(db_file) as pipe:
        try:
            rows = pipe.execute(q)
        except sqlite3.OperationalError:
            return logger.exception("Failed to execute the query, Please check once, ")

        result = []
        for row in rows.fetchall()[:20]:
            result.append([])
            for cell in row:
                result[-1].append(cell if cell is not None else "")

        column_names = [description[0] for description in rows.description]
        if result:
            print(
                tabulate.tabulate(
                    result,
                    headers=column_names,
                    tablefmt="rounded_grid",
                    maxheadercolwidths=12,
                    maxcolwidths=12,
                )
            )
        secho(f"Returned {len(result)} rows.", fg="green")


@faq.command(short_help="fetches the timestamp of the latest run")
@option(
    "-p",
    "--allow-pending",
    default=False,
    show_default=True,
    help="consider runs, whose status are pending",
    is_flag=True,
)
@pass_context
def latest_run(ctx: Context, allow_pending: bool):
    ctx.invoke(
        query,
        collection_path=ctx.parent.params["collection_path"],
        q=(
            "SELECT PROJECTNAME, STRFTIME('%d/%m/%Y, %H:%M', STARTED, 'localtime') as STARTED, "
            "STRFTIME('%d/%m/%Y, %H:%M', ENDED, 'localtime') as ENDED, TESTID"
            " FROM RUNBASE WHERE ENDED <> '' ORDER BY RUNBASE.STARTED DESC LIMIT 1"
            if not allow_pending
            else "SELECT PROJECTNAME, STRFTIME('%d/%m/%Y, %H:%M', STARTED, 'localtime') as STARTED,"
            " STRFTIME('%d/%m/%Y, %H:%M', ENDED, 'localtime') as ENDED,"
            " TESTID FROM RUNBASE ORDER BY RUNBASE.STARTED DESC LIMIT 1"
        ),
    )


@faq.command(
    short_help="fetches the number of yet to export task",
    help="returns list of tasks of form: (ticket_id, task_type, dropped_date, is_picked, test_id)",
)
@pass_context
def yet_to_process(ctx: Context):
    ctx.invoke(
        query,
        collection_path=ctx.parent.params["collection_path"],
        q=(
            "SELECT ticketID, type, STRFTIME('%d/%m/%Y, %H:%M', dropped, 'localtime'),"
            " picked, test_id FROM TASKBASE WHERE "
            "PROCESSED = 0"
        ),
    )



from concurrent.futures import ThreadPoolExecutor, Future
from pathlib import Path
from tortoise import run_async
from tortoise.connection import connections
from handshake.services.DBService.lifecycle import (
    init_tortoise_orm,
    db_path,
    attachment_folder,
)
from handshake.services.DBService.migrator import migration
from loguru import logger
from shutil import unpack_archive, rmtree, move, copytree
from tempfile import mkdtemp
from typing import Tuple, List
from sqlite3.dbapi2 import Connection, connect


async def reset_sqlite_sequence(output_db_path):
    await init_tortoise_orm(
        output_db_path, True, init_script=True, close_it=True, avoid_config=True
    )
    connection = connections.get("default")
    await connection.execute_query("delete from sqlite_sequence")
    await connection.close()


def prep_minis(zipped_results: Path) -> Path:
    logger.debug("preparing a temp. copy of {}...", zipped_results.name)
    temp_folder = Path(mkdtemp(prefix="handshake-merge-"))
    is_zip = zipped_results.is_file()

    try:
        if is_zip:
            logger.debug("unpacking {} into {}", zipped_results, temp_folder)
            temp_folder /= zipped_results.name.split(".")[0]
            temp_folder.mkdir()
            unpack_archive(zipped_results, temp_folder, "bztar")
            logger.debug("checking for possible migration for {}", db_path(temp_folder))
            logger.debug("running migrator on {}", db_path(temp_folder))
        else:
            logger.debug("copying provided folder {} to a temp folder", zipped_results)
            temp_folder /= zipped_results.name
            copytree(zipped_results, temp_folder)
            logger.debug("{} is now copied to {}", zipped_results, temp_folder)
            logger.debug("running migrator on {}", db_path(temp_folder))
    except Exception as error:
        logger.warning("cleaning temp folders as it failed.")
        rmtree(temp_folder.parent)
        raise error

    migration(db_path(temp_folder))
    return temp_folder


class Merger:
    avoid_tables = {"configbase", "sqlite_sequence"}

    def __init__(self, output_folder):
        output_folder = output_folder
        self.output_db_path = db_path(output_folder)

        set_default_first = not Path(output_folder).exists()
        if set_default_first:
            Path(output_folder).mkdir(exist_ok=True)

        run_async(
            reset_sqlite_sequence(
                self.output_db_path,
            )
        )

    def start(self, to_merge: Tuple[str]):
        futures: List[Future[Path[str]]] = []
        with ThreadPoolExecutor(max_workers=6) as prep:
            for collection in to_merge:
                futures.append(prep.submit(prep_minis, Path(collection)))

        paths: List[Path[str]] = []

        for future in futures:
            has_failed = future.exception()
            if has_failed:
                logger.error(
                    "Failed to create temporary directory for merging refer to the error: {}. hence skipping",
                    has_failed,
                )
                continue

            path = future.result()
            paths.append(path)

        try:
            self.merge_internals(paths)

            dest = attachment_folder(self.output_db_path)

            with ThreadPoolExecutor(max_workers=6) as after_merge:
                logger.debug("moving static folders and files")
                for collection in paths:
                    for test_run in attachment_folder(
                        db_path(str(collection))
                    ).iterdir():
                        after_merge.submit(move, test_run, dest)

            logger.debug("done, cleaning temp folders...")

        finally:
            with ThreadPoolExecutor(max_workers=6) as cleaner:
                for collection in paths:
                    cleaner.submit(rmtree, collection.parent)
            logger.debug("temp folders have been removed successfully")
        logger.info("Merge Completed, saved to {}!", self.output_db_path.parent)

    def merge_internals(self, paths: List[Path]):
        with connect(self.output_db_path) as connection:
            connection.isolation_level = None
            for path in paths:
                self.merge_database(connection, db_path(path))

    def merge_database(self, connection: Connection, child_db_path: Path):
        connection.execute("ATTACH ? as dba", (str(child_db_path.absolute()),))
        connection.execute("BEGIN;")
        logger.debug("Appending records")

        try:
            self.merge_tables(connection, child_db_path)
        except Exception as error:
            logger.exception(
                "Failed to merge with {}. hence skipping it. due to {}",
                child_db_path.parent.name,
                repr(error),
            )
            connection.rollback()
        else:
            connection.commit()
            logger.debug(
                "Merged {} with output db successfully.",
                child_db_path.parent.name,
            )

        finally:
            connection.execute("detach dba;")

    def merge_tables(self, connection: Connection, child_path: Path):
        for row in connection.execute(
            "SELECT * FROM dba.sqlite_master WHERE type='table'"
        ):
            table_name = row[1]

            if table_name in self.avoid_tables:
                continue

            cols = set(
                [
                    _[0]
                    for _ in connection.execute(
                        f"SELECT name from PRAGMA_TABLE_INFO('{table_name}')"
                    )
                ]
            )

            if "id" in cols:
                cols.remove("id")

            joined_cols = ", ".join(cols)
            # join cols even if id is not in the cols, since migration scripts can rearrange
            # the order of the columns in the table
            combine = f"INSERT INTO {table_name}({joined_cols}) SELECT {joined_cols} FROM dba.{table_name}"

            logger.debug("Executed, {} on {}", combine, child_path.parent.name)
            connection.execute(combine)



from handshake.services.SchedularService.constants import (
    EXPORT_RUN_PAGE_FILE_NAME,
    EXPORT_PROJECTS_FILE_NAME,
    EXPORT_RUNS_PAGE_FILE_NAME,
    EXPORT_OVERVIEW_PAGE,
    EXPORT_ALL_SUITES,
    EXPORT_SUITE_PAGE_FILE_NAME,
    EXPORT_SUITE_TESTS_PAGE,
    EXPORT_TEST_ASSERTIONS,
    EXPORT_TEST_ENTITY_ATTACHMENTS,
    EXPORT_SUITE_RETRIED_MAP,
    exportAttachmentFolderName,
)
from handshake.Exporters.exporter import Exporter
from loguru import logger
from asyncio import gather, TaskGroup
from shutil import rmtree
from pathlib import Path
from json import dumps
from aiofiles.os import mkdir
from aiofiles import open


async def write_as_plain_string(path, string):
    async with open(path, "w") as file_pointer:
        await file_pointer.write(string)


class JsonExporter(Exporter):
    def __init__(self, db_path: Path, save_in: Path, dev_run: bool = False):
        super().__init__(dev_run)
        self.save_in = save_in / exportAttachmentFolderName
        self.db_path: Path = db_path

    def prepare(self):
        # we reset entire export folder
        if self.save_in.exists():
            logger.debug("removing previous results")
            rmtree(self.save_in)

        self.save_in.parent.mkdir(exist_ok=True)
        self.save_in.mkdir(exist_ok=False)

    def completed(self):
        logger.info("Export Completed, saved in {}", self.save_in)

    async def export_test_run_summary(self, test_id: str, summary):
        await mkdir(self.save_in / str(test_id))
        await write_as_plain_string(
            self.save_in / str(test_id) / EXPORT_RUN_PAGE_FILE_NAME,
            dumps(summary),
        )

    async def export_project_summary(self, run_feed, projects_feed):
        return gather(
            write_as_plain_string(
                self.save_in / EXPORT_RUNS_PAGE_FILE_NAME,
                dumps(run_feed),
            ),
            write_as_plain_string(
                self.save_in / EXPORT_PROJECTS_FILE_NAME,
                dumps(projects_feed),
            ),
        )

    async def export_overview_of_test_run(self, run_id, summary):
        await write_as_plain_string(
            self.save_in / run_id / EXPORT_OVERVIEW_PAGE,
            dumps(summary),
        )

    async def export_all_suites_of_test_run(self, run_id: str, all_suites):
        async with TaskGroup() as prep:
            for suite in all_suites:
                suite_id = suite["suiteID"]
                await mkdir(self.save_in / run_id / suite_id)

                prep.create_task(
                    write_as_plain_string(
                        self.save_in / run_id / suite_id / EXPORT_SUITE_PAGE_FILE_NAME,
                        dumps(suite),
                    )
                )

        await write_as_plain_string(
            self.save_in / run_id / EXPORT_ALL_SUITES, dumps(all_suites)
        )

    async def export_tests(self, run_id, suite_id, tests):
        await write_as_plain_string(
            self.save_in / run_id / suite_id / EXPORT_SUITE_TESTS_PAGE,
            dumps(tests),
        )

    async def export_attachments(
        self, run_id, suite_id, assertion_records, written_records
    ):
        return gather(
            write_as_plain_string(
                self.save_in / run_id / suite_id / EXPORT_TEST_ENTITY_ATTACHMENTS,
                dumps(dict(written=written_records)),
            ),
            write_as_plain_string(
                self.save_in / run_id / suite_id / EXPORT_TEST_ASSERTIONS,
                dumps(assertion_records),
            ),
        )

    async def export_retries_map(self, run_id, suite_id, retried_map):
        await write_as_plain_string(
            self.save_in / run_id / suite_id / EXPORT_SUITE_RETRIED_MAP,
            dumps(retried_map),
        )



from typing import Optional, Any, Union, Tuple
from handshake.services.SchedularService.constants import exportExportFileName, JobType
from handshake.services.SchedularService.register import warn_about_test_run
from handshake.services.DBService.models.dynamic_base import TaskBase
from handshake.services.DBService.models.result_base import (
    SuiteBase,
    SuiteType,
)
from handshake.services.DBService.models.attachmentBase import AssertBase
from handshake.services.DBService.models.static_base import StaticBase
from handshake.Exporters.exporter import Exporter
from handshake.services.DBService.lifecycle import attachment_folder
from loguru import logger
from pathlib import Path
from aiofiles.os import mkdir
from datetime import datetime
from tortoise.expressions import Q

try:
    from openpyxl import load_workbook, Workbook
    from openpyxl.worksheet.worksheet import Worksheet
    from openpyxl.cell import Cell
    from openpyxl.styles.alignment import Alignment
    from openpyxl.comments.comments import Comment
    from openpyxl.worksheet.table import Table, TableStyleInfo
    from openpyxl.formatting.formatting import ConditionalFormatting
    from openpyxl.styles.fills import (
        FILL_SOLID,
        PatternFill,
        FILL_PATTERN_DARKHORIZONTAL,
    )

    excel_export = True
except ImportError:
    Workbook = Any
    Worksheet = Any
    ConditionalFormatting = Any
    excel_export = False
    Cell = Any

# PLEASE DO NOT TYPECAST EXCEL RELATED THINGS UNLESS WE HANDLE IT IN EXCEPTION


def save_datetime_in_excel(obj: datetime, cell: Optional[Cell] = None):
    full_text = obj.strftime("%Y-%m-%d %I:%M:%S %p")
    if not cell:
        return full_text
    cell.comment = Comment(full_text, "Handshake", width=150, height=40)
    return obj.strftime("%I:%M:%S %p")


def calc_percentage(num, deno):
    return (num / deno) if deno else 0


class ExcelExporter(Exporter):
    template: Workbook
    standing_format: ConditionalFormatting
    percentage_format: ConditionalFormatting
    parent_links = {}
    test_link = {}

    test_summary_row = 3
    test_summary_col = 4
    test_summary_files_row = -1

    def __init__(self, db_path: Path, dev_run: bool = False):
        super().__init__(dev_run)
        self.save_in = attachment_folder(db_path)

    def prepare(self):
        self.template = load_workbook(Path(__file__).parent / "Template.xlsx")

    async def start_exporting(
        self,
        run_id: Optional[str] = None,
        skip_project_summary: bool = False,
        skip_prepare: bool = False,
    ):
        if not excel_export:
            return logger.error(
                "Please install excel exporter using"
                " pip install handshake[excel-export] or poetry add handshake[excel-export]"
            )
        task = await TaskBase.filter(type=JobType.EXPORT_EXCEL, test_id=run_id).first()
        try:
            await super().start_exporting(run_id, True)
        except Exception as error:
            logger.exception("Failed to export this test run in excel due to {}", error)
            await warn_about_test_run(
                run_id,
                "Failed to export this test run in excel",
                task.type,
                error=repr(error),
                task=task.ticketID,
            )
            task.processed = True
            await task.save()
            return None

        self.save_in.mkdir(exist_ok=True)
        if not run_id:
            return None

        not (self.save_in / str(run_id)).exists() and await mkdir(
            self.save_in / str(run_id)
        )
        self.template.save(self.save_in / str(run_id) / exportExportFileName)

        task.processed = True
        task.picked = False
        await task.save()
        return None

    def completed(self):
        logger.info(
            "Excel is available in Attachments folder, Please check at the {}",
            self.save_in,
        )

    async def export_test_run_summary(self, test_id: str, summary):
        index_sheet = self.template["Index"]

        edit_cell(
            index_sheet,
            self.test_summary_row,
            self.test_summary_col,
            summary["projectName"],
            True,
        )
        self.test_summary_row += 1

        standing_cell = edit_cell(
            index_sheet,
            self.test_summary_row,
            self.test_summary_col,
            summary["standing"].lower().capitalize(),
            True,
        )
        self.test_summary_row += 1

        self.standing_format, self.percentage_format, *_ = list(
            index_sheet.conditional_formatting
        )

        standing_cell.comment = Comment(
            f"Run Status: {summary['status'].lower().capitalize()}\n"
            f"Exit Code: {summary['exitCode']}",
            summary["framework"],
            width=200,
        )

        edit_cell(
            index_sheet,
            self.test_summary_row,
            self.test_summary_col,
            format_duration(summary["duration"] / 1000),
        )
        self.test_summary_row += 1

        edit_cell(
            index_sheet,
            self.test_summary_row,
            self.test_summary_col,
            summary["suites"],
        )
        self.test_summary_row += 1

        edit_cell(
            index_sheet,
            self.test_summary_row,
            self.test_summary_col,
            summary["tests"],
        )
        self.test_summary_row += 1

        for status in ("passed", "failed", "skipped", "xfailed", "xpassed"):
            edit_cell(
                index_sheet,
                self.test_summary_row,
                self.test_summary_col,
                summary[status],
            )
            self.test_summary_row += 1  # at the end of the loop for overall percentage

        edit_cell(
            index_sheet,
            self.test_summary_row,
            self.test_summary_col,
            calc_percentage(
                summary["passedSuites"] + summary["skippedSuites"],
                summary["suites"],
            ),
        )
        self.test_summary_row += 2  # files, platform
        self.test_summary_files_row = self.test_summary_row - 1

        edit_cell(
            index_sheet,
            self.test_summary_row,
            self.test_summary_col,
            summary["platform"],
            True,
        )
        self.test_summary_row += 1

        edit_cell(
            index_sheet,
            self.test_summary_row,
            self.test_summary_col,
            summary["framework"],
        )
        # starting with the timeline table
        self.test_summary_row += 3
        edit_cell(
            index_sheet,
            self.test_summary_row,
            self.test_summary_col,
            save_datetime_in_excel(datetime.fromisoformat(summary["started"])),
        )
        self.test_summary_row += 1

        edit_cell(
            index_sheet,
            self.test_summary_row,
            self.test_summary_col,
            save_datetime_in_excel(datetime.fromisoformat(summary["ended"])),
        )
        self.test_summary_row += 1

        edit_cell(
            index_sheet,
            self.test_summary_row,
            self.test_summary_col,
            save_datetime_in_excel(datetime.now()),
            resize=True,
        )
        self.test_summary_row += 1

    async def export_run_page(self, run_id: str, skip_recent_suites: bool = False):
        await super().export_run_page(run_id, True)

    async def export_project_summary(self, run_feed, projects_feed): ...

    async def export_overview_of_test_run(self, run_id: str, summary):
        index_sheet = self.template["Index"]

        edit_cell(
            index_sheet,
            self.test_summary_files_row,
            self.test_summary_col,
            summary["aggregated"]["files"],
        )

    async def export_all_suites(self, run_id: str, export_suite_wise: bool = True):
        await super().export_all_suites(run_id, False)

    async def export_all_suites_of_test_run(self, run_id, all_suites):
        if not all_suites:
            return
        suites_sheet = self.template["Test Scenarios"]
        table_style = TableStyleInfo(name="TableStyleMedium23", showRowStripes=True)

        columns = (
            "title",
            "aliasID",
            "total_duration",
            "description",
            "standing",
            "tests",
            "rollup_passed",
            "rollup_failed",
            "rollup_skipped",
            "rollup_xfailed",
            "rollup_xpassed",
            "numberOfErrors",
            "error",
            "started",
            "ended",
            "parent",
            "simplified",
            "file",
            "duration",
            "setup_duration",
            "teardown_duration",
            "retried_later",
            "Parent Alias",
        )
        alias = {
            "aliasID": "Alias",
            "numberOfErrors": "Errors",
            "duration": "test_duration",
            "tests": "passed%",
            "simplified": "Ran with",
            "standing": "status",
            "rollup_passed": "Passed",
            "rollup_failed": "Failed",
            "rollup_skipped": "Skipped",
            "rollup_xfailed": "XFailed",
            "rollup_xpassed": "XPassed",
            "retried_later": "Retried Later",
            "setup_duration": "Setup",
            "teardown_duration": "Teardown",
        }

        suites_sheet.freeze_panes = "B1"  # freeze first col (title)

        for row_index, suite in enumerate(all_suites):
            for header_index, header in enumerate(columns):
                if not row_index:
                    suites_sheet.cell(1, header_index + 1).value = alias.get(
                        header, header
                    ).capitalize()

                cell = suites_sheet.cell(row_index + 2, header_index + 1)
                cell.alignment = Alignment("right")
                match header:
                    case "Parent Alias":
                        value: Union[Any, Tuple[str, str, str], bool] = (
                            self.parent_links.get(suite["parent"], False)
                        )
                    case _:
                        value = suite[header]

                to_save = value
                resize = True
                match header:
                    case "started":
                        to_save = save_datetime_in_excel(
                            datetime.fromisoformat(value), cell
                        )
                    case "ended":
                        to_save = save_datetime_in_excel(
                            datetime.fromisoformat(value), cell
                        )
                    case "duration":
                        to_save = format_duration(value / 1000)
                        resize = (
                            False  # else it considers the length of the formula string
                        )
                        resize_col(suites_sheet, cell, 10)
                    case "total_duration":
                        to_save = format_duration(value / 1000)
                        resize = False
                        resize_col(suites_sheet, cell, 6)
                    case "setup_duration":
                        to_save = format_duration(value / 1000)
                        resize = False
                        resize_col(suites_sheet, cell, 6)
                    case "teardown_duration":
                        to_save = format_duration(value / 1000)
                        resize = False
                        resize_col(suites_sheet, cell, 7)
                    case "tests":
                        to_save = calc_percentage(
                            suite["rollup_passed"] + suite["rollup_skipped"],
                            suite["rollup_tests"],
                        )
                        cell.number_format = "0%"
                        copy_format_to_cell(suites_sheet, cell, self.percentage_format)
                        resize = False
                    case "title":
                        resize = True
                        self.parent_links[suite["suiteID"]] = (
                            f"#'{suites_sheet.title}'!{cell.column_letter}{cell.row}",
                            value,
                            suite["aliasID"],
                        )
                    case "standing":
                        to_save = value.lower().capitalize()
                        resize = True
                        copy_format_to_cell(suites_sheet, cell, self.standing_format)
                    case "title":
                        cell.alignment = Alignment(wrapText=True)
                        resize_col(suites_sheet, cell, 30)
                        resize = False
                    case "description":
                        if value:
                            cell.alignment = Alignment(wrapText=True)
                            resize_col(suites_sheet, cell, 30)
                            resize = False
                    case "error":
                        if value:
                            resize_col(suites_sheet, cell, 30)
                            resize = False
                    case "parent":
                        to_save = "〰️"
                        cell.alignment = Alignment("center", "center")
                        if value and self.parent_links[value]:
                            to_save = f"{suite['parent_title']} 🔗"
                            cell.hyperlink = self.parent_links[value][0]
                            resize = True
                            add_comment_to_cell(cell, self.parent_links[value][1])
                    case "Parent Alias":
                        to_save = "〰️"
                        cell.alignment = Alignment("center", "center")
                        if value:
                            to_save = f"{value[-1]} 🔗"
                            cell.hyperlink = value[0]
                            resize = True
                            add_comment_to_cell(cell, value[-1])
                    case "retried_later":
                        to_save = "YES" if value else "NO"
                    case "file":
                        add_comment_to_cell(cell, value)
                        resize = False
                        resize_col(suites_sheet, cell, 10)
                    case _:
                        to_save = value
                        resize = False
                        resize_col(suites_sheet, cell, 6)

                edit_cell(
                    suites_sheet, row_index + 2, header_index + 1, to_save, resize
                )

        suites_sheet.add_table(
            Table(
                ref=f"A1:{suites_sheet.cell(1, 1).offset(0, len(columns) - 1).column_letter}"
                f"{len(all_suites) + 1}",
                displayName="Test_Scenarios",
                tableStyleInfo=table_style,
            )
        )

    @staticmethod
    def export_test_query(based_on: str):
        return (
            SuiteBase.filter(
                ~Q(suiteType=SuiteType.SUITE) & Q(session__test_id=based_on)
            ),
            AssertBase.filter(entity__session__test_id=based_on),
            StaticBase.filter(entity__session__test_id=based_on),
        )

    async def export_tests(self, run_id, suite_id, tests):
        if not tests:
            return
        test_sheet = self.template["Test Cases"]
        hooks_sheet = self.template["Hooks"]
        table_style = TableStyleInfo(name="TableStyleMedium23", showRowStripes=True)

        columns = (
            "title",
            "aliasID",
            "total_duration",
            "suiteType",
            "description",
            "standing",
            "assertions",
            "numberOfErrors",
            "error",
            "started",
            "ended",
            "parent",
            "file",
            "duration",
            "setup_duration",
            "teardown_duration",
            "retried_later",
            "Parent Alias",
        )
        alias = {
            "aliasID": "alias",
            "duration": "test_duration",
            "numberOfErrors": "Errors",
            "simplified": "Ran with",
            "standing": "status",
            "retried_later": "Retried Later",
            "setup_duration": "Setup",
            "teardown_duration": "Teardown",
            "suiteType": "Type",
        }

        test_sheet.freeze_panes = "B1"  # freeze first col (title)
        hooks_sheet.freeze_panes = "B1"  # freeze first col (title)

        test_rows = 2
        hooks_row = 2

        for row_index, test in enumerate(tests):
            is_test = test["suiteType"] == SuiteType.TEST

            for header_index, header in enumerate(columns):
                if not row_index:
                    header_value = alias.get(header, header).capitalize()
                    test_sheet.cell(1, header_index + 1).value = header_value
                    hooks_sheet.cell(1, header_index + 1).value = header_value

                sheet = test_sheet if is_test else hooks_sheet
                cell = sheet.cell(test_rows if is_test else hooks_row, header_index + 1)
                cell.alignment = Alignment("right")

                match header:
                    case "Parent Alias":
                        value: Union[Any, Tuple[str, str, str], bool] = (
                            self.parent_links.get(test["parent"], False)
                        )
                    case _:
                        value = test[header]

                resize = True
                to_save = value

                match header:
                    case "started":
                        to_save = save_datetime_in_excel(
                            datetime.fromisoformat(value), cell
                        )
                    case "ended":
                        to_save = save_datetime_in_excel(
                            datetime.fromisoformat(value), cell
                        )
                    case "duration":
                        to_save = format_duration(value / 1000)
                        resize = (
                            False  # else it considers the length of the formula string
                        )
                        resize_col(sheet, cell, 10)
                    case "total_duration":
                        to_save = format_duration(value / 1000)
                        resize = False
                        resize_col(sheet, cell, 6)
                    case "setup_duration":
                        to_save = format_duration(value / 1000)
                        resize = False
                        resize_col(sheet, cell, 6)
                    case "teardown_duration":
                        to_save = format_duration(value / 1000)
                        resize = False
                        resize_col(sheet, cell, 7)
                    case "title":
                        resize = True
                        self.parent_links[test["suiteID"]] = (
                            f"#'{test_sheet.title}'!{cell.column_letter}{cell.row}",
                            value,
                            test["aliasID"],
                        )
                    case "standing":
                        to_save = value.lower().capitalize()
                        resize = True
                        copy_format_to_cell(sheet, cell, self.standing_format)
                    case "title":
                        cell.alignment = Alignment(wrapText=True)
                        resize_col(sheet, cell, 30)
                        resize = False
                    case "description":
                        if value:
                            cell.alignment = Alignment(wrapText=True)
                            resize_col(sheet, cell, 30)
                            resize = False
                    case "error":
                        if value:
                            resize_col(sheet, cell, 30)
                            resize = False
                    case "parent":
                        to_save = ""
                        index = (
                            test_rows
                            if test["suiteType"] == SuiteType.SETUP
                            else (test_rows - 1)
                        )
                        link = self.parent_links.get(
                            value,
                            (f"#'{test_sheet.title}'!A{index}", "", ""),
                        )
                        if value and link:
                            to_save = f"{test['parent_title']} 🔗"
                            resize = True
                            cell.hyperlink = link[0]
                            cell.alignment = Alignment("center", "center")

                            if link[1]:
                                add_comment_to_cell(cell, link[1])
                    case "Parent Alias":
                        to_save = ""
                        index = (
                            test_rows
                            if test["suiteType"] == SuiteType.SETUP
                            else (test_rows - 1)
                        )
                        link = (
                            value
                            if value
                            else (f"#'{test_sheet.title}'!A{index}", "", "")
                        )
                        if value and link:
                            to_save = f"{link[2]} 🔗"
                            resize = True
                            cell.hyperlink = link[0]
                            cell.alignment = Alignment("center", "center")

                            if link[2]:
                                add_comment_to_cell(cell, link[2])
                    case "retried_later":
                        to_save = "YES" if value else "NO"
                    case "file":
                        add_comment_to_cell(cell, value)
                        resize = False
                        resize_col(sheet, cell, 10)
                    case _:
                        to_save = value
                        resize = False
                        resize_col(sheet, cell, 6)

                edit_cell(
                    sheet,
                    cell.row,
                    cell.column,
                    to_save,
                    resize,
                )
            if is_test:
                test_rows += 1
            else:
                hooks_row += 1

        test_rows > 2 and test_sheet.add_table(
            Table(
                ref=f"A1:{test_sheet.cell(1, 1).offset(0, len(columns) - 1).column_letter}"
                f"{test_rows - 1}",
                displayName="Test_Cases",
                tableStyleInfo=table_style,
            )
        )
        hooks_row > 2 and hooks_sheet.add_table(
            Table(
                ref=f"A1:{hooks_sheet.cell(1, 1).offset(0, len(columns) - 1).column_letter}"
                f"{hooks_row - 1}",
                displayName="Hooks",
                tableStyleInfo=table_style,
            )
        )

    async def export_retries_map(self, run_id, suite_id, retried_map): ...

    async def export_attachments(
        self, run_id, suite_id, assertion_records, written_records
    ):
        if not assertion_records:
            return
        assertion_sheet = self.template["Assertions"]
        table_style = TableStyleInfo(name="TableStyleMedium23", showRowStripes=True)

        columns = (
            "title",
            "raw",
            "entity_id",
            "passed",
            "interval",
            "wait",
            "Test Alias",
        )
        alias = {"entity_id": "entity", "raw": "description"}

        assertion_sheet.freeze_panes = "B1"  # freeze first col (title)
        fill = PatternFill(
            start_color="FFC7CE", end_color="FFC7CE", fill_type=FILL_SOLID
        )
        assertions = [_ for __ in assertion_records for _ in assertion_records[__]]

        for row_index, assertion in enumerate(assertions):
            for header_index, header in enumerate(columns):
                if not row_index:
                    assertion_sheet.cell(1, header_index + 1).value = alias.get(
                        header, header
                    ).capitalize()

                cell = assertion_sheet.cell(row_index + 2, header_index + 1)
                cell.alignment = Alignment("right")
                if not assertion["passed"]:
                    cell.fill = fill

                match header:
                    case "Test Alias":
                        value = self.parent_links.get(assertion["entity_id"], False)
                    case _:
                        value = assertion[header]
                to_save = value
                resize = True
                match header:
                    case "title":
                        cell.alignment = Alignment(wrapText=True)
                        resize_col(assertion_sheet, cell, 30)
                        resize = False
                    case "raw":
                        if value:
                            cell.alignment = Alignment(wrapText=True)
                            resize_col(assertion_sheet, cell, 30)
                            resize = False
                    case "entity_id":
                        to_save = "〰️"
                        cell.alignment = Alignment("center", "center")
                        link, tip, parent_alias = self.parent_links.get(
                            value, ("", "", "")
                        )
                        if value and link:
                            to_save = f"{tip} 🔗"
                            cell.hyperlink = link
                            resize = True
                            add_comment_to_cell(cell, tip)
                    case "Test Alias":
                        to_save = "〰️"
                        cell.alignment = Alignment("center", "center")
                        link, tip, parent_alias = value
                        if value and link:
                            to_save = f"{parent_alias} 🔗"
                            cell.hyperlink = link
                            resize = True
                            add_comment_to_cell(cell, parent_alias)
                    case "passed":
                        to_save = "YES" if value else "NO"
                    case _:
                        to_save = value
                        resize = False
                        resize_col(assertion_sheet, cell, 6)

                edit_cell(
                    assertion_sheet, row_index + 2, header_index + 1, to_save, resize
                )

        assertion_sheet.add_table(
            Table(
                ref=f"A1:{assertion_sheet.cell(1, 1).offset(0, len(columns) - 1).column_letter}"
                f"{len(assertions) + 1}",
                displayName="Assertions",
                tableStyleInfo=table_style,
            )
        )


def edit_cell(
    sheet: Worksheet, row: int, col: int, value: str, resize: Optional[bool] = False
):
    cell = sheet.cell(row, col)
    cell.value = value
    if value and resize:
        resize_col(sheet, cell, len(value))
    return cell


def resize_col(sheet: Worksheet, cell: Cell, length: int):
    sheet.column_dimensions[cell.column_letter].width = max(
        sheet.column_dimensions[cell.column_letter].width, length
    )


def format_duration(value: int):
    return f"""=IF({value}>=3600, TEXT(INT({value}/3600), "0") & " hrs " & 
                TEXT(INT(MOD({value}, 3600)/60), "0") & " mins " & TEXT(MOD({value}, 60), "0") & " secs",
                IF({value}>=60, TEXT(INT({value}/60), "0") & " mins " & TEXT(MOD({value}, 60), "0") & " secs",
                IF({value}>=1, TEXT({value}, "0") & " secs", TEXT({value}*1000, "0") & " ms")))"""


def copy_format_to_cell(
    sheet: Worksheet, cell: Cell, formatting_rules: ConditionalFormatting
):
    for rule in formatting_rules.rules:
        sheet.conditional_formatting.add(
            f"{cell.column_letter}{cell.row}",
            rule,
        )


def add_comment_to_cell(cell: Cell, value: str):
    cell.comment = Comment(value, "Handshake")
    cell.comment.width = 150
    cell.comment.height = 30 + (len(value) / 30) * 30



from handshake.services.DBService.models.attachmentBase import TestLogBase, LogType
from handshake.services.DBService.models.dynamic_base import TaskBase
from loguru import logger
from typing import Optional
from tortoise.expressions import Q, Subquery
from typing import Union
from uuid import UUID


async def pruneTasks(task_id: Optional[str] = ""):
    if task_id:
        logger.warning("Deleting Few Tasks as per request")
    else:
        logger.debug("Pruning some old tasks")

    to_prune = await TaskBase.filter(
        Q(
            test_id__in=Subquery(
                TestLogBase.filter(type=LogType.ERROR)
                .filter(
                    test_id__in=Subquery(
                        TaskBase.filter(processed=False)
                        .only("test_id")
                        .distinct()
                        .values_list("test_id", flat=True)
                    )
                )
                .only("test_id")
                .distinct()
                .values_list("test_id", flat=True)
            )
        )
    )

    for task in to_prune:
        logger.error(
            "Found an error in this task: {}. hence marking it as processed."
            " Please report it as an issue if it was not expected.",
            task.ticketID,
        )
        task.processed = True
        task.picked = True

    if not to_prune:
        logger.debug("No Tasks were pruned.")
    else:
        await TaskBase.bulk_update(
            to_prune,
            (
                "picked",
                "processed",
            ),
            100,
        )

    if task_id:
        # removing the prune task
        await TaskBase.filter(ticketID=task_id).delete()


# async def skipTestRunForNow():
#     logger.warning(
#         "Marking few tasks to be picked in the next iteration since the details were not updated yet, which are "
#         "related to error test runs"
#     )
#
#     requests = (
#         await TaskBase.filter(type=JobType.SKIP_REQUEST)
#         .distinct()
#         .values_list("ticketID", flat=True)
#     )
#     to_skip = await TaskBase.filter(
#         test_id__in=Subquery(
#             TaskBase.filter(type=JobType.SKIP_REQUEST)
#             .distinct()
#             .values_list("test_id", flat=True)
#         )
#     )
#
#     for task in to_skip:
#         logger.error(
#             "Skipping this: {}. will be picked in the next export command"
#             " Please report it as an issue if it was not expected.",
#             task.ticketID,
#         )
#         task.skip = True
#         task.skip_times += 1
#         task.picked = False
#
#     await TaskBase.bulk_update(
#         to_skip,
#         ("skipped", "picked"),
#         100,
#     )
#
#     not to_skip and logger.debug("No Tasks were skipped.")
#
#     # removing the skip request task
#     requests and await TaskBase.filter(ticketID__in=requests).delete()



from importlib.util import find_spec

if not find_spec("sanic_ext"):

    def definition(**_):
        def accept(func):
            return func

        return accept

else:
    from sanic_ext import openapi

    definition = openapi.definition



from handshake.services.DBService.models.result_base import (
    SessionBase,
    RunBase,
    SuiteBase,
)
from handshake.services.DBService.models.config_base import TestConfigBase
from uuid import UUID
from typing import Union
from traceback import format_exc
from handshake.services.SchedularService.register import (
    cancel_patch_for_test_run,
)
from handshake.services.SchedularService.constants import JobType
from handshake.services.DBService.models.dynamic_base import TaskBase
from handshake.services.DBService.models.types import Status, SuiteType
from handshake.services.SchedularService.refer_types import PathTree, PathItem
from handshake.services.SchedularService.modifySuites import fetch_key_from_status
from tortoise.functions import Sum, Max, Min, Count, Lower, Coalesce
from datetime import datetime, timezone
from typing import List, Tuple
from pathlib import Path
from os.path import join
from loguru import logger
from tortoise.expressions import Q, Subquery
from typing import Optional
from asyncio import gather


class PatchTestRun:
    actual_start = "actual_start"
    actual_end = "actual_end"

    def __init__(self, test_id):
        self.test: Optional[RunBase] = None
        self.test_id = test_id
        self.task: Optional[TaskBase] = None

    async def fetch_records(self):
        self.test = await RunBase.filter(testID=self.test_id).first()
        self.task = await TaskBase.filter(ticketID=self.test_id).first()

    async def patch_test(self):
        await self.fetch_records()
        if not await self.do_we_need_patch():
            return False
        await self.patch_test_values()
        await self.mark_processed()
        return True

    async def mark_processed(self):
        self.task.processed = True
        await self.task.save()

    async def fetch_suite_summary(self):
        test_config = await TestConfigBase.filter(test_id=self.test_id).first()

        refer = SuiteBase

        if test_config and test_config.avoidParentSuitesInCount:
            # consider cucumber files
            # if the feature file has 3 scenarios, then, if the user sets the avoidParentSuitesInCount,
            # we would get a value of 3 as total scenarios else 4
            refer = SuiteBase.filter(~Q(parent=""))

        # we want to count the amount of suite status
        summary = dict(passed=0, failed=0, skipped=0, xpassed=0, xfailed=0)
        summary.update(
            await refer.filter(
                Q(session__test_id=self.test_id)
                & Q(suiteType=SuiteType.SUITE)
                & ~Q(standing=Status.RETRIED)
            )
            .annotate(count=Count("suiteID"), status=Lower("standing"))
            .group_by("standing")
            .values_list("status", "count")
        )
        summary["count"] = sum(summary.values())

        return summary

    async def fetch_agg_values_of_test(self):
        filtered_sessions = SessionBase.filter(test_id=self.test_id)

        filtered = SuiteBase.filter(
            Q(session_id__in=Subquery(filtered_sessions.values("sessionID")))
            & Q(suiteType=SuiteType.TEST)
            & Q(retried_later=False)
        )

        # we get the values of the test run from the non-retried test entities
        test_result = (
            await filtered.annotate(
                passed=Coalesce(Sum("passed"), 0),
                failed=Coalesce(Sum("failed"), 0),
                skipped=Coalesce(Sum("skipped"), 0),
                tests=Coalesce(Sum("tests"), 0),
                xfailed=Coalesce(Sum("xfailed"), 0),
                xpassed=Coalesce(Sum("xpassed"), 0),
            )
            .first()
            .values("passed", "failed", "skipped", "tests", "xpassed", "xfailed")
        )
        test_result.update(
            await filtered_sessions.annotate(
                actual_end=Max("ended"),
                actual_start=Min("started"),
            )
            .first()
            .values(
                self.actual_start,
                self.actual_end,
            )
        )

        return test_result

    async def patch_test_values(self):
        (summary, test_result, retried_suites) = await gather(
            self.fetch_suite_summary(),
            self.fetch_agg_values_of_test(),
            # just to get the count of the retried suites we had in this run
            SuiteBase.filter(Q(parent="") & Q(standing=Status.RETRIED)).count(),
        )

        # the start date was initially when we start the shipment
        # now it is when the first session starts
        started = (
            test_result.get(self.actual_start, self.test.started) or self.test.started
        )
        ended = test_result.get(
            self.actual_end, datetime.now(timezone.utc)
        ) or datetime.now(timezone.utc)
        test_result.pop(self.actual_start) if self.actual_start in test_result else ...
        test_result.pop(self.actual_end) if self.actual_end in test_result else ...

        await self.test.update_from_dict(
            dict(
                **{_: test_result[_] or 0 for _ in test_result.keys()},
                **{
                    f"{_}Suites" if _ != "count" else "suites": summary[_] or 0
                    for _ in summary.keys()
                },
                retried=retried_suites,
                started=started,
                ended=ended,
                duration=(ended - started).total_seconds() * 1000,
                specStructure=simplify_file_paths(
                    [
                        path
                        for path in await SuiteBase.filter(
                            Q(session__test_id=self.test_id)
                            & Q(suiteType=SuiteType.SUITE)
                            & ~Q(standing=Status.RETRIED)
                            & Q(parent="")
                        )
                        .group_by("file")
                        .prefetch_related("rollup")
                        .annotate(tests=Sum("rollup__tests"))
                        .values_list("file", "tests")
                    ]
                ),
                standing=fetch_key_from_status(
                    summary["passed"],
                    summary["failed"],
                    summary["skipped"],
                    summary["xfailed"],
                    summary["xpassed"],
                ),
            )
        )
        await self.test.save()

    async def pick_it_later(self):
        self.task.picked = False
        await self.task.save()

    async def do_we_need_patch(self):
        # check: 1
        # to see if the test is not pending
        if self.test.standing != Status.PENDING:
            logger.warning("{} was already processed", self.test_id)
            await self.mark_processed()
            return False

        # check: 2
        # check if any child suites are yet to be processed,
        # this should not happen, as we are exporting all the related suites before test run

        pending_items = (
            await SuiteBase.filter(session__test_id=self.test_id)
            .filter(
                Q(standing=Status.YET_TO_CALCULATE)
                | Q(standing=Status.PENDING)
                | Q(standing=Status.PROCESSING)
            )
            .count()
        )

        if pending_items > 0:
            await skip_coz_error(
                self.test_id,
                f"Failed to export the Test Run: {self.test_id} of project: {self.test.projectName},"
                f" because some of the suites were not processed",
                incomplete=self.test.projectName,
                pending_suites=pending_items,
            )
            return False

        return True


async def skip_coz_error(test_id: Union[str, UUID], reason: str, **extra) -> False:
    return await cancel_patch_for_test_run(
        test_id,
        reason,
        False,
        JobType.MODIFY_TEST_RUN,
        type=JobType.MODIFY_TEST_RUN,
        **extra,
    )


def simplify_file_paths(paths: List[Tuple[str, int]]):
    tree: PathTree = {"current": "", "paths": {}, "suites": 1}
    _paths: List[PathItem] = [
        dict(children=list(reversed(Path(path[0]).parts)), pointer=tree, count=path[1])
        for path in paths
    ]

    # Tree - Builder
    while _paths:
        path_to_include = _paths[-1]
        if not path_to_include["children"]:
            _paths.pop()
            path_to_include["pointer"]["suites"] = path_to_include["count"]
            continue

        parent_pointer: PathTree = path_to_include["pointer"]
        possible_parent: str = path_to_include["children"].pop()
        # that dict will be inside the tree

        # NOTE: we are not calculating the cumulative count of suites for the folder
        pointing_to = parent_pointer["paths"].setdefault(
            possible_parent,
            {
                "current": join(parent_pointer["current"], possible_parent),
                "paths": {},
                "suites": 1,
            },
        )
        path_to_include["pointer"] = pointing_to

    # Reducer
    while True:
        stack = [(_, tree) for _ in tree["paths"].keys()]
        movements = 0

        while stack:
            node_key, parent_node = stack.pop()
            target_node = parent_node["paths"][node_key]

            children = list(target_node["paths"].keys())
            if len(children) > 1:
                for child_key in target_node["paths"]:
                    stack.append((child_key, target_node))
                continue

            if not children:
                continue
            child_key = children.pop()

            movements += 1
            target_popped = parent_node["paths"].pop(node_key)
            new_key = join(node_key, child_key)
            parent_node["paths"][new_key] = target_popped["paths"].pop(child_key)
            new_child = parent_node["paths"][new_key]
            new_child["current"] = join(target_popped["current"], child_key)

            for child_key in new_child["paths"]:
                stack.append((child_key, new_child))

        if not movements:
            break

    return tree["paths"]


async def patchTestRun(test_id: str):
    patcher = PatchTestRun(test_id)
    to_return = False

    try:
        if await patcher.patch_test():
            logger.debug("Completed the export for test run: {}", test_id)
            to_return = True
    except Exception:
        await skip_coz_error(
            test_id,
            f"Failed to export the test run, error in calculation, {format_exc()}",
        )
    return to_return



from enum import StrEnum, IntEnum


class MigrationStatus(StrEnum):
    PASSED = "PASSED"
    PENDING = "PENDING"
    FAILED = "FAILED"


class LogGeneratedBy(IntEnum):
    USER = 0
    API = 1
    SCHEDULER = 2


class MigrationTrigger(StrEnum):
    AUTOMATIC = "AUTO"
    CLI = "CLI"


class Status(StrEnum):
    PASSED = "PASSED"
    PENDING = "PENDING"
    FAILED = "FAILED"
    SKIPPED = "SKIPPED"
    YET_TO_CALCULATE = (
        "YET_TO_CALC"  # needs to be updated by our server to either passed or failed
    )
    # yet_to_calc is mostly seen for the suite
    RETRIED = "RETRIED"
    PROCESSING = "PROCESSING"
    XPASSED = "XPASSED"
    XFAILED = "XFAILED"


class RunStatus(StrEnum):
    COMPLETED = "COMPLETED"
    PENDING = "PENDING"
    INTERRUPTED = "INTERRUPTED"
    INTERNAL_ERROR = "INTERNAL_ERROR"
    SKIPPED = "SKIPPED"  # not in use as of now
    EXPECTED_TO_FAIL = "EXPECTED_TO_FAIL"


class SuiteType(StrEnum):
    TEST = "TEST"
    SUITE = "SUITE"
    SETUP = "SETUP"
    TEARDOWN = "TEARDOWN"


class LogType(StrEnum):
    ERROR = "ERROR"
    WARN = "WARN"
    INFO = "INFO"


class AttachmentType(StrEnum):
    LABEL = "LABEL"
    ENV = "ENV"
    CONFIG = "CONFIG"
    VIDEO = "VIDEO"
    PNG = "PNG"
    DESC = "DESC"
    IMG = "IMAGE"
    LINK = "LINK"
    ASSERT = "ASSERT"
    NOTE = "NOTE"  # depreciated
    LOG = "LOG"


class LogDisplayType(StrEnum):
    SEPARATE = "SEPARATE"
    TOGETHER = "TOGETHER"


class ConfigKeys(StrEnum):
    maxRunsPerProject = "MAX_RUNS_PER_PROJECT"
    version = "VERSION"
    recentlyDeleted = "RECENTLY_DELETED"
    reset_test_run = "RESET_FIX_TEST_RUN"



from handshake.services.Endpoints.core import service_provider
from handshake.services.DBService.lifecycle import init_tortoise_orm, close_connection
from handshake.services.DBService.shared import set_test_id


@service_provider.before_server_start
async def before_start_of_day(*args):
    set_test_id()
    await init_tortoise_orm(avoid_config=True)


@service_provider.after_server_stop
async def close_app(*args):
    await close_connection()
